/*
Copyright (C) 2017-2018 Keziah Wesley

You can redistribute and/or modify this file under the terms of the
GNU Affero General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any
later version.

This file is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public
License along with this file. If not, see
<http://www.gnu.org/licenses/>.
*/

#include <assert.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

enum {
  PT_UNBOUND,
  PT_FIX = (1 << 16),
  PT_MASK = (0x7fff << 16),
};

enum {
  TT_UNBOUND = PT_UNBOUND,
  TT_FIX = PT_FIX,
};

enum {
  B_IF,
  B_IF_NOT,
};

enum {
  R_STACK = 0xff,
};

typedef struct {
  uint32_t type;
  uint32_t _stuff;
  union {
    int64_t fix;
  };
} object;

inline static int64_t GETFIX (object o) {
  assert ((o.type & PT_MASK) == PT_FIX);
  return o.fix;
}
inline static object NEWFIX (int64_t v) {
  return (object){ .type = TT_FIX, .fix = (v) };
}
#define RESULT(i) (&bp[(p[i] == R_STACK) ? sp++ : p[i]])
#define JUMP(off) do { p = (off < 0) ? oldp + off : p + off; } while (0)

int snprint_op_disasm (char ** buf, size_t outlen, const uint8_t ** p, size_t inlen);
void run (const uint8_t * op, const uint8_t * p, object * bp, int sp);

static void trace(const uint8_t * p0, const uint8_t * p)
{
  char buf[512];
  char *b = buf;
  b += snprintf(buf, 512, "%ld: ", p-p0);
  int n = snprint_op_disasm (&b, 512, &p, (size_t)-1);
  assert (!n);
  printf(buf);
}

enum {
  EOKAY,
  EBADOPCODE,
  EBADOPARGS,
};

#define TRACE(p0, p) trace(p0, p)

#include "gen.c"

int main ()
{
  // hand-assembled iterative factorial:
  uint8_t locl_N = 0;
  uint8_t locl_TMP1 = 1;
  uint8_t progn[] = {
    I_GRTRB, locl_N, 1, B_IF, 1,             //      <GRTR? N 1 + LOOP>
    I_HALT,                                  //      <HALT>
    I_MUL, locl_N, locl_TMP1, locl_TMP1,     // LOOP <MUL N TMP1 = TMP1>
    I_SUBB, locl_N, 1, locl_N,               //      <SUB N 1 = N>
    I_GRTRB, locl_N, 1, B_IF, (uint8_t)-8,   //      <GRTR? N 1 + LOOP>
    I_HALT,                                  //      <HALT>
  };

  object stack[128];
  memset(stack, '\0', 128 * sizeof(*stack));
  stack[0] = NEWFIX(7);
  stack[1] = NEWFIX(1);

  run (progn, progn, stack, 2);
  printf("%lu", GETFIX(stack[1]));
  return 0;
}
